import json
import struct
import traceback


class Channel(object):
    """
    Channel is a higher level abstraction over a low-level bytestream.

    You can send and receive JSON serializable object directly with this interface

    For now this class does not do much, but we could imagine some sort compression or other
    transformation being added here
    """

    def __init__(self, stream):
        self._stream = stream
        self._fmt = struct.Struct("<I")

    def send(self, obj):
        try:
            to_send = json.dumps(obj, ensure_ascii=False, separators=(",", ":")).encode(
                "utf-8"
            )
            sz = len(to_send)
            self._stream.write(self._fmt.pack(sz))
            self._stream.write(to_send)
        except EOFError as e:
            raise RuntimeError("Cannot send object over streaming interface: %s" % e)
        except BaseException as e:
            raise ValueError("Cannot serialize object: %s" % traceback.format_exc())

    def recv(self, timeout=None):
        # To receive, we first receive the size of the object and then the object itself
        try:
            sz_bytes = self._stream.read(self._fmt.size, timeout)
            msg_sz = self._fmt.unpack(sz_bytes)[0]
            obj_bytes = self._stream.read(msg_sz, timeout)
            return json.loads(obj_bytes)
        except EOFError as e:
            raise RuntimeError("Cannot receive object over streaming interface: %s" % e)
        except BaseException as e:
            raise ValueError("Cannot deserialize object: %s" % traceback.format_exc())

    def fileno(self):
        return self._stream.fileno()



import json
from .card import MetaflowCard, MetaflowCardComponent
from .renderer_tools import render_safely


class TestStringComponent(MetaflowCardComponent):
    REALTIME_UPDATABLE = True

    def __init__(self, text):
        self._text = text

    def render(self):
        return str(self._text)

    def update(self, text):
        self._text = text


class TestPathSpecCard(MetaflowCard):
    type = "test_pathspec_card"

    def render(self, task):
        import random
        import string

        return "%s %s" % (
            task.pathspec,
            "".join(
                random.choice(string.ascii_uppercase + string.digits) for _ in range(6)
            ),
        )


class TestEditableCard(MetaflowCard):
    type = "test_editable_card"

    separator = "$&#!!@*"

    ALLOW_USER_COMPONENTS = True

    def __init__(self, components=[], **kwargs):
        self._components = components

    def render(self, task):
        return self.separator.join([str(comp) for comp in self._components])


class TestEditableCard2(MetaflowCard):
    type = "test_editable_card_2"

    separator = "$&#!!@*"

    ALLOW_USER_COMPONENTS = True

    def __init__(self, components=[], **kwargs):
        self._components = components

    def render(self, task):
        return self.separator.join([str(comp) for comp in self._components])


class TestNonEditableCard(MetaflowCard):
    type = "test_non_editable_card"

    separator = "$&#!!@*"

    def __init__(self, components=[], **kwargs):
        self._components = components

    def render(self, task):
        return self.separator.join([str(comp) for comp in self._components])


class TestMockCard(MetaflowCard):
    type = "test_mock_card"

    def __init__(self, options={"key": "dummy_key"}, **kwargs):
        self._key = options["key"]

    def render(self, task):
        task_data = task[self._key].data
        return "%s" % task_data


class TestErrorCard(MetaflowCard):
    type = "test_error_card"

    # the render function will raise Exception
    def render(self, task):
        raise Exception("Unknown Things Happened")


class TestTimeoutCard(MetaflowCard):
    type = "test_timeout_card"

    def __init__(self, options={"timeout": 50}, **kwargs):
        super().__init__()
        self._timeout = 10
        if "timeout" in options:
            self._timeout = options["timeout"]

    # the render function will raise Exception
    def render(self, task):
        import time

        time.sleep(self._timeout)
        return "%s" % task.pathspec


REFRESHABLE_HTML_TEMPLATE = """
<html>
<script> 
var METAFLOW_RELOAD_TOKEN = "[METAFLOW_RELOAD_TOKEN]"

window.metaflow_card_update = function(data) {
    document.querySelector("h1").innerHTML = JSON.stringify(data);
}
</script>
<h1>[PATHSPEC]</h1>
<h1>[REPLACE_CONTENT_HERE]</h1>
</html>
"""


class TestJSONComponent(MetaflowCardComponent):

    REALTIME_UPDATABLE = True

    def __init__(self, data):
        self._data = data

    @render_safely
    def render(self):
        return self._data

    def update(self, data):
        self._data = data


class TestRefreshCard(MetaflowCard):
    """
    This card takes no components and helps test the `current.card.refresh(data)` interface.
    """

    HTML_TEMPLATE = REFRESHABLE_HTML_TEMPLATE

    RUNTIME_UPDATABLE = True

    ALLOW_USER_COMPONENTS = True

    # Not implementing Reload Policy here since the reload Policy is set to always
    RELOAD_POLICY = MetaflowCard.RELOAD_POLICY_ALWAYS

    type = "test_refresh_card"

    def render(self, task) -> str:
        return self._render_func(task, self.runtime_data)

    def _render_func(self, task, data):
        return self.HTML_TEMPLATE.replace(
            "[REPLACE_CONTENT_HERE]", json.dumps(data["user"])
        ).replace("[PATHSPEC]", task.pathspec)

    def render_runtime(self, task, data):
        return self._render_func(task, data)

    def refresh(self, task, data):
        return data


import hashlib


def _component_values_to_hash(components):
    comma_str = ",".join(["".join(x) for v in components.values() for x in v])
    return hashlib.sha256(comma_str.encode("utf-8")).hexdigest()


class TestRefreshComponentCard(MetaflowCard):
    """
    This card takes components and helps test the `current.card.components["A"].update()`
    interface
    """

    HTML_TEMPLATE = REFRESHABLE_HTML_TEMPLATE

    RUNTIME_UPDATABLE = True

    ALLOW_USER_COMPONENTS = True

    # Not implementing Reload Policy here since the reload Policy is set to always
    RELOAD_POLICY = MetaflowCard.RELOAD_POLICY_ONCHANGE

    type = "test_component_refresh_card"

    def __init__(self, components=[], **kwargs):
        self._components = components

    def render(self, task) -> str:
        # Calling `render`/`render_runtime` wont require the `data` object
        return self.HTML_TEMPLATE.replace(
            "[REPLACE_CONTENT_HERE]", json.dumps(self._components)
        ).replace("[PATHSPEC]", task.pathspec)

    def render_runtime(self, task, data):
        return self.render(task)

    def refresh(self, task, data):
        # Govers the information passed in the data update
        return data["components"]

    def reload_content_token(self, task, data):
        if task.finished:
            return "final"
        return "runtime-%s" % _component_values_to_hash(data["components"])


class TestImageCard(MetaflowCard):
    """Card that renders a tiny PNG using ``TaskToDict.parse_image``."""

    type = "test_image_card"

    def render(self, task):
        from .convert_to_native_type import TaskToDict
        import base64

        png_bytes = base64.b64decode(
            "iVBORw0KGgoAAAANSUhEUgAAAAEAAAABCAYAAAAfFcSJAAAADUlEQVR4nGNgYGBgAAAABQABRDE8UwAAAABJRU5ErkJggg=="
        )
        img_src = TaskToDict().parse_image(png_bytes)
        return f"<html><img src='{img_src}' /></html>"



from functools import partial
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Union,
)

from metaflow.debug import debug
from metaflow.exception import MetaflowException

from .user_step_decorator import StepMutator, UserStepDecoratorBase

if TYPE_CHECKING:
    import metaflow.decorators
    import metaflow.flowspec


class MutableStep:
    IGNORE = 1
    ERROR = 2
    OVERRIDE = 3

    def __init__(
        self,
        flow_spec: "metaflow.flowspec.FlowSpec",
        step: Union[
            Callable[["metaflow.decorators.FlowSpecDerived"], None],
            Callable[["metaflow.decorators.FlowSpecDerived", Any], None],
        ],
        pre_mutate: bool = False,
        statically_defined: bool = False,
        inserted_by: Optional[str] = None,
    ):
        from .mutable_flow import MutableFlow

        self._mutable_container = MutableFlow(
            flow_spec,
            pre_mutate=pre_mutate,
            statically_defined=statically_defined,
            inserted_by=inserted_by,
        )
        self._flow_cls = flow_spec.__class__
        self._my_step = step
        self._pre_mutate = pre_mutate
        self._statically_defined = statically_defined
        self._inserted_by = inserted_by
        if self._inserted_by is None:
            # This is an error because MutableSteps should only be created by
            # StepMutators or FlowMutators. We need to catch it now because otherwise
            # we may put stuff on the command line (with --with) that would get added
            # twice and weird behavior may ensue.
            raise MetaflowException(
                "MutableStep should only be created by StepMutators or FlowMutators. "
                "This is an internal error."
            )

    @property
    def flow(self) -> "metaflow.user_decorator.mutable_flow.MutableFlow":
        """
        The flow that contains this step

        Returns
        -------
        MutableFlow
            The flow that contains this step
        """
        return self._mutable_container

    @property
    def decorator_specs(
        self,
    ) -> Generator[Tuple[str, str, List[Any], Dict[str, Any]], None, None]:
        """
        Iterate over all the decorator specifications of this step. Note that the same
        type of decorator may be present multiple times and no order is guaranteed.

        The returned tuple contains:
        - The decorator's name (shortest possible)
        - The decorator's fully qualified name (in the case of Metaflow decorators, this
          will indicate which extension the decorator comes from)
        - A list of positional arguments
        - A dictionary of keyword arguments

        You can use the resulting tuple to remove a decorator for example

        Yields
        ------
        str, str, List[Any], Dict[str, Any]
            A tuple containing the decorator name, it's fully qualified name,
            a list of positional arguments, and a dictionary of keyword arguments.
        """
        for deco in self._my_step.decorators:
            # 3.7 does not support yield foo, *bar syntax so we
            # work around
            r = [
                deco.name,
                "%s.%s"
                % (
                    deco.__class__.__module__,
                    deco.__class__.__name__,
                ),
            ]
            r.extend(deco.get_args_kwargs())
            yield tuple(r)

        for deco in self._my_step.wrappers:
            r = [
                UserStepDecoratorBase.get_decorator_name(deco.__class__),
                deco.decorator_name,
            ]
            r.extend(deco.get_args_kwargs())
            yield tuple(r)

        for deco in self._my_step.config_decorators:
            r = [
                UserStepDecoratorBase.get_decorator_name(deco.__class__),
                deco.decorator_name,
            ]
            r.extend(deco.get_args_kwargs())
            yield tuple(r)

    def add_decorator(
        self,
        deco_type: Union[partial, UserStepDecoratorBase, str],
        deco_args: Optional[List[Any]] = None,
        deco_kwargs: Optional[Dict[str, Any]] = None,
        duplicates: int = IGNORE,
    ) -> None:
        """
        Add a Metaflow step-decorator or a user step-decorator to a step.

        You can either add the decorator itself or its decorator specification for it
        (the same you would get back from decorator_specs). You can also mix and match
        but you cannot provide arguments both through the string and the
        deco_args/deco_kwargs.

        As an example:
        ```
        from metaflow import environment
        ...
        my_step.add_decorator(environment, deco_kwargs={"vars": {"foo": 42})}
        ```

        is equivalent to:
        ```
        my_step.add_decorator('environment:vars={"foo": 42}')
        ```

        is equivalent to:
        ```
        my_step.add_decorator('environment', deco_kwargs={"vars":{"foo": 42}})
        ```

        but this is not allowed:
        ```
        my_step.add_decorator('environment:vars={"bar" 43}', deco_kwargs={"vars":{"foo": 42}})
        ```

        Note in the case where you specify a
        string for the decorator, there is no need to import the decorator.

        The string syntax is useful to, for example, allow decorators to be stored as
        strings in a configuration file.

        You can only add StepMutators in the pre_mutate stage.

        In terms of precedence for decorators:
          - if a decorator can be applied multiple times (like `@card`) all decorators
            added are kept.
          - if `duplicates` is set to `MutableStep.IGNORE`, then the decorator
            being added is ignored (in other words, the existing decorator has precedence).
          - if `duplicates` is set to `MutableStep.OVERRIDE`, then the *existing*
            decorator is removed and this newly added one replaces it (in other
            words, the newly added decorator has precedence).
          - if `duplicates` is set to `MutableStep.ERROR`, then an error is raised but only
            if the newly added decorator is *static* (ie: defined directly in the code).
            If not, it is ignored.

        Parameters
        ----------
        deco_type : Union[partial, UserStepDecoratorBase, str]
            The decorator class to add to this step.
        deco_args : List[Any], optional, default None
            Positional arguments to pass to the decorator.
        deco_kwargs : Dict[str, Any], optional, default None
            Keyword arguments to pass to the decorator.
        duplicates : int, default MutableStep.IGNORE
            Instruction on how to handle duplicates. It can be one of:
            - `MutableStep.IGNORE`: Ignore the decorator if it already exists.
            - `MutableStep.ERROR`: Raise an error if the decorator already exists.
            - `MutableStep.OVERRIDE`: Remove the existing decorator and add this one.
        """
        # Prevent circular import
        from metaflow.decorators import (
            DuplicateStepDecoratorException,
            StepDecorator,
            extract_step_decorator_from_decospec,
        )

        deco_args = deco_args or []
        deco_kwargs = deco_kwargs or {}

        def _add_step_decorator(step_deco):
            if deco_args:
                raise MetaflowException(
                    "Step decorators do not take additional positional arguments"
                )
            # Update kwargs:
            step_deco.attributes.update(deco_kwargs)

            # Check duplicates
            def _do_add():
                step_deco.statically_defined = self._statically_defined
                step_deco.inserted_by = self._inserted_by
                self._my_step.decorators.append(step_deco)
                debug.userconf_exec(
                    "Mutable step adding step decorator '%s' to step '%s'"
                    % (deco_type, self._my_step.name)
                )

            existing_deco = [
                d for d in self._my_step.decorators if d.name == step_deco.name
            ]

            if step_deco.allow_multiple or not existing_deco:
                _do_add()
            elif duplicates == MutableStep.IGNORE:
                # If we ignore, we do not add the decorator
                debug.userconf_exec(
                    "Mutable step ignoring step decorator '%s' on step '%s' "
                    "(already exists and duplicates are ignored)"
                    % (step_deco.name, self._my_step.name)
                )
            elif duplicates == MutableStep.OVERRIDE:
                # If we override, we remove the existing decorator and add this one
                debug.userconf_exec(
                    "Mutable step overriding step decorator '%s' on step '%s' "
                    "(removing existing decorator and adding new one)"
                    % (step_deco.name, self._my_step.name)
                )
                self._my_step.decorators = [
                    d for d in self._my_step.decorators if d.name != step_deco.name
                ]
                _do_add()
            elif duplicates == MutableStep.ERROR:
                # If we error, we raise an exception
                if self._statically_defined:
                    raise DuplicateStepDecoratorException(step_deco.name, self._my_step)
                else:
                    debug.userconf_exec(
                        "Mutable step ignoring step decorator '%s' on step '%s' "
                        "(already exists and non statically defined)"
                        % (step_deco.name, self._my_step.name)
                    )
            else:
                raise ValueError("Invalid duplicates value: %s" % duplicates)

        if isinstance(deco_type, str):
            step_deco, has_args_kwargs = extract_step_decorator_from_decospec(deco_type)
            if (deco_args or deco_kwargs) and has_args_kwargs:
                raise MetaflowException(
                    "Cannot specify additional arguments when adding a user step "
                    "decorator using a decospec that already has arguments"
                )

            if isinstance(step_deco, StepDecorator):
                _add_step_decorator(step_deco)
            else:
                # User defined decorator.
                if not self._pre_mutate and isinstance(step_deco, StepMutator):
                    raise MetaflowException(
                        "Adding step mutator '%s' from %s is only allowed in the "
                        "`pre_mutate` method and not the `mutate` method"
                        % (step_deco.decorator_name, self._inserted_by)
                    )

                if deco_args or deco_kwargs:
                    # We need to recreate the object if there were args or kwargs
                    # since they were not in the string
                    step_deco = step_deco.__class__(*deco_args, **deco_kwargs)

                step_deco.add_or_raise(
                    self._my_step,
                    self._statically_defined,
                    duplicates,
                    self._inserted_by,
                )
            return

        if isinstance(deco_type, type) and issubclass(deco_type, UserStepDecoratorBase):
            # We can only add step mutators in the pre mutate stage.
            if not self._pre_mutate and issubclass(deco_type, StepMutator):
                raise MetaflowException(
                    "Adding step mutator '%s' from %s is only allowed in the "
                    "`pre_mutate` method and not the `mutate` method"
                    % (step_deco.decorator_name, self._inserted_by)
                )
            debug.userconf_exec(
                "Mutable step adding decorator %s to step %s"
                % (deco_type, self._my_step.name)
            )

            d = deco_type(*deco_args, **deco_kwargs)
            # add_or_raise properly registers the decorator
            d.add_or_raise(
                self._my_step, self._statically_defined, duplicates, self._inserted_by
            )
            return

        # At this point, it should be a regular Metaflow step decorator
        if (
            not isinstance(deco_type, partial)
            or len(deco_type.args) != 1
            or not issubclass(deco_type.args[0], StepDecorator)
        ):
            raise TypeError(
                "add_decorator takes a metaflow decorator or user StepDecorator"
            )

        deco_type = deco_type.args[0]
        _add_step_decorator(
            deco_type(
                attributes=deco_kwargs,
                statically_defined=self._statically_defined,
                inserted_by=self._inserted_by,
            )
        )

    def remove_decorator(
        self,
        deco_name: str,
        deco_args: Optional[List[Any]] = None,
        deco_kwargs: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """
        Remove a step-level decorator. To remove a decorator, you can pass the decorator
        specification (obtained from `decorator_specs` for example).
        Note that if multiple decorators share the same decorator specification
        (very rare), they will all be removed.

        You can only remove StepMutators in the `pre_mutate` method.

        Parameters
        ----------
        deco_name : str
            Decorator specification of the decorator to remove. If nothing else is
            specified, all decorators matching that name will be removed.
        deco_args : List[Any], optional, default None
            Positional arguments to match the decorator specification.
        deco_kwargs : Dict[str, Any], optional, default None
            Keyword arguments to match the decorator specification.

        Returns
        -------
        bool
            Returns True if a decorator was removed.
        """

        do_all = deco_args is None and deco_kwargs is None
        did_remove = False
        canonical_deco_type = UserStepDecoratorBase.get_decorator_by_name(deco_name)
        if issubclass(canonical_deco_type, UserStepDecoratorBase):
            for attr in ["config_decorators", "wrappers"]:
                new_deco_list = []
                for deco in getattr(self._my_step, attr):
                    if deco.decorator_name == canonical_deco_type.decorator_name:
                        if do_all:
                            continue  # We remove all decorators with this name
                        if deco.get_args_kwargs() == (
                            deco_args or [],
                            deco_kwargs or {},
                        ):
                            if not self._pre_mutate and isinstance(deco, StepMutator):
                                raise MetaflowException(
                                    "Removing step mutator '%s' from %s is only allowed in the "
                                    "`pre_mutate` method and not the `mutate` method"
                                    % (deco.decorator_name, self._inserted_by)
                                )
                            did_remove = True
                            debug.userconf_exec(
                                "Mutable step removing user step decorator '%s' from step '%s'"
                                % (deco.decorator_name, self._my_step.name)
                            )
                        else:
                            new_deco_list.append(deco)
                    else:
                        new_deco_list.append(deco)
                setattr(self._my_step, attr, new_deco_list)

        if did_remove:
            return True
        new_deco_list = []
        for deco in self._my_step.decorators:
            if deco.name == deco_name:
                if do_all:
                    continue  # We remove all decorators with this name
                # Check if the decorator specification matches
                if deco.get_args_kwargs() == (deco_args, deco_kwargs):
                    did_remove = True
                    debug.userconf_exec(
                        "Mutable step removing step decorator '%s' from step '%s'"
                        % (deco.name, self._my_step.name)
                    )
                else:
                    new_deco_list.append(deco)
            else:
                new_deco_list.append(deco)

        self._my_step.decorators = new_deco_list

        if did_remove:
            return True

        debug.userconf_exec(
            "Mutable step did not find decorator '%s' to remove from step '%s'"
            % (deco_name, self._my_step.name)
        )
        return False



from .card_modules import MetaflowCardComponent
from .card_modules.card import create_component_id
from .card_modules.basic import ErrorComponent, SectionComponent
from .card_modules.components import (
    UserComponent,
    StubComponent,
)
from .exception import ComponentOverwriteNotSupportedException
from metaflow.metaflow_config import RUNTIME_CARD_RENDER_INTERVAL
import uuid
import json
import platform
from collections import OrderedDict
import time

_TYPE = type


def get_card_class(card_type):
    from metaflow.plugins import CARDS

    filtered_cards = [card for card in CARDS if card.type == card_type]
    if len(filtered_cards) == 0:
        return None
    return filtered_cards[0]


def _component_is_valid(component):
    """
    Validates if the component is of the correct class.
    """
    return issubclass(type(component), MetaflowCardComponent)


def warning_message(message, logger=None, ts=False):
    if logger:
        msg = "[@card WARNING] %s" % message
        logger(msg, timestamp=ts, bad=True)


class WarningComponent(ErrorComponent):
    def __init__(self, warning_message):
        super().__init__("@card WARNING", warning_message)


class ComponentStore:
    """
    The `ComponentStore` object helps store the components for a single card in memory.
    This class has combination of a array/dictionary like interfaces to access/change the stored components.

    It exposes the `append` /`extend` methods (like an array) to add components.
    It also exposes the `__getitem__`/`__setitem__` methods (like a dictionary) to access the components by their Ids.
    """

    def _set_component_map(self):
        """
        The `_component_map` attribute is supposed to be a dictionary so that we can access the components by their ids.
        But we also want to maintain order in which components are inserted since all of these components are going to be visible on a UI.
        Since python3.6 dictionaries are ordered by default so we can use the default python `dict`.
        """
        self._component_map = {}

    def __init__(self, logger, card_type=None, components=None, user_set_id=None):
        self._logger = logger
        self._card_type = card_type
        self._user_set_id = user_set_id
        self._layout_last_changed_on = time.time()
        self._set_component_map()
        if components is not None:
            for c in list(components):
                self._store_component(c, component_id=None)

    @property
    def layout_last_changed_on(self):
        """This property helps the CardComponentManager identify when the layout of the card has changed so that it can trigger a re-render of the card."""
        return self._layout_last_changed_on

    def _realtime_updateable_components(self):
        for c in self._component_map.values():
            if c.REALTIME_UPDATABLE:
                yield c

    def _store_component(self, component, component_id=None):
        if not _component_is_valid(component):
            warning_message(
                "Component (%s) is not a valid MetaflowCardComponent. It will not be stored."
                % str(component),
                self._logger,
            )
            return
        if component_id is not None:
            component.component_id = component_id
        elif component.component_id is None:
            component.component_id = create_component_id(component)
        setattr(component, "_logger", self._logger)
        self._component_map[component.component_id] = component
        self._layout_last_changed_on = time.time()

    def _remove_component(self, component_id):
        del self._component_map[component_id]
        self._layout_last_changed_on = time.time()

    def __iter__(self):
        return iter(self._component_map.values())

    def __setitem__(self, key, value):
        if self._component_map.get(key) is not None:
            # This is the equivalent of calling `current.card.components["mycomponent"] = Markdown("## New Component")`
            # We don't support the replacement of individual components in the card.
            # Instead we support rewriting the entire component array instead.
            # So users can run `current.card[ID] = [FirstComponent, SecondComponent]` which will instantiate an entirely
            # new ComponentStore.
            # So we should throw an error over here, since it is clearly an operation which is not supported.
            raise ComponentOverwriteNotSupportedException(
                key, self._user_set_id, self._card_type
            )
        else:
            self._store_component(value, component_id=key)

    def __getitem__(self, key):
        if key not in self._component_map:
            # Store a stub-component in place since `key` doesnt exist.
            # If the user does a `current.card.append(component, id=key)`
            # then the stub component will be replaced by the actual component.
            self._store_component(StubComponent(key), component_id=key)
        return self._component_map[key]

    def __delitem__(self, key):
        if key not in self._component_map:
            raise KeyError(
                "MetaflowCardComponent with id `%s` not found. Available components for the cards include : %s"
                % (key, ", ".join(self.keys()))
            )
        self._remove_component(key)

    def __contains__(self, key):
        return key in self._component_map

    def append(self, component, id=None):
        self._store_component(component, component_id=id)

    def extend(self, components):
        for c in components:
            self._store_component(c, component_id=None)

    def clear(self):
        self._component_map.clear()

    def keys(self):
        return list(self._component_map.keys())

    def values(self):
        return self._component_map.values()

    def __str__(self):
        return "Card components present in the card: `%s` " % ("`, `".join(self.keys()))

    def __len__(self):
        return len(self._component_map)


def _object_is_json_serializable(obj):
    try:
        json.dumps(obj)
        return True
    except TypeError as e:
        return False


class CardComponentManager:
    """
    This class manages the card's state for a single card.
    - It uses the `ComponentStore` to manage the storage of the components
    - It exposes methods to add, remove and access the components.
    - It exposes a `refresh` method that will allow refreshing a card with new data
    for realtime(ish) updates.
    - The `CardComponentCollector` exposes convenience methods similar to this class for a default
    editable card. These methods include :
        - `append`
        - `extend`
        - `clear`
        - `refresh`
        - `components`
        - `__iter__`

    ## Usage Patterns :

    ```python
    current.card["mycardid"].append(component, id="comp123")
    current.card["mycardid"].extend([component])
    current.card["mycardid"].refresh(data) # refreshes the card with new data
    current.card["mycardid"].components["comp123"] # returns the component with id "comp123"
    current.card["mycardid"].components["comp123"].update()
    current.card["mycardid"].components.clear() # Wipe all the components
    del current.card["mycardid"].components["mycomponentid"] # Delete a component
    ```
    """

    def __init__(
        self,
        card_uuid,
        decorator_attributes,
        card_creator,
        components=None,
        logger=None,
        no_warnings=False,
        user_set_card_id=None,
        runtime_card=False,
        card_options=None,
        refresh_interval=5,
    ):
        self._card_creator_args = dict(
            card_uuid=card_uuid,
            user_set_card_id=user_set_card_id,
            runtime_card=runtime_card,
            decorator_attributes=decorator_attributes,
            card_options=card_options,
            logger=logger,
        )
        self._card_creator = card_creator
        self._refresh_interval = refresh_interval
        self._last_layout_change = None
        self._latest_user_data = None
        self._last_refresh = 0
        self._last_render = 0
        self._render_seq = 0
        self._logger = logger
        self._no_warnings = no_warnings
        self._warn_once = {
            "update": {},
            "not_implemented": {},
        }
        card_type = decorator_attributes["type"]

        if components is None:
            self._components = ComponentStore(
                logger=self._logger,
                card_type=card_type,
                user_set_id=user_set_card_id,
                components=None,
            )
        else:
            self._components = ComponentStore(
                logger=self._logger,
                card_type=card_type,
                user_set_id=user_set_card_id,
                components=list(components),
            )

    def append(self, component, id=None):
        self._components.append(component, id=id)

    def extend(self, components):
        self._components.extend(components)

    def clear(self):
        self._components.clear()

    def _card_proc(self, mode, sync=False):
        self._card_creator.create(**self._card_creator_args, mode=mode, sync=sync)

    def refresh(self, data=None, force=False):
        self._latest_user_data = data
        nu = time.time()
        first_render = True if self._last_render == 0 else False

        if nu - self._last_refresh < self._refresh_interval:
            # rate limit refreshes: silently ignore requests that
            # happen too frequently
            return
        self._last_refresh = nu

        # This block of code will render the card in `render_runtime` mode when:
        # 1. refresh is called with `force=True`
        # 2. Layout of the components in the card has changed. i.e. The actual elements in the component array have changed.
        # 3. The last time the card was rendered was more the minimum interval after which they should be rendered.
        last_rendered_before_minimum_interval = (
            nu - self._last_refresh
        ) > RUNTIME_CARD_RENDER_INTERVAL
        layout_has_changed = (
            self._last_layout_change != self.components.layout_last_changed_on
            or self._last_layout_change is None
        )
        if force or last_rendered_before_minimum_interval or layout_has_changed:
            self._render_seq += 1
            self._last_render = nu
            self._card_proc("render_runtime")
            # The below `if not first_render` condition is a special case for the following scenario:
            # Lets assume the case that the user is only doing `current.card.append` followed by `refresh`.
            # In this case, there will be no process executed in `refresh` mode since `layout_has_changed`
            # will always be true and as a result there will be no data update that informs the UI of the RELOAD_TOKEN change.
            # This will cause the UI to seek for the data update object but will constantly find None. So if it is not
            # the first render then we should also have a `refresh` call followed by a `render_runtime` call so
            # that the UI can always be updated with the latest data.
            if not first_render:
                # For the general case, the CardCreator's ProcessManager run's the `refresh` / `render_runtime` in a asynchronous manner.
                # Due to this when the `render_runtime` call is happening, an immediately subsequent call to `refresh` will not be able to
                # execute since the card-process manager will be busy executing the `render_runtime` call and ignore the `refresh` call.
                # Hence we need to pass the `sync=True` argument to the `refresh` call so that the `refresh` call is executed synchronously and waits for the
                # `render_runtime` call to finish.
                self._card_proc("refresh", sync=True)
            # We set self._last_layout_change so that when self._last_layout_change is not the same
            # as `self.components.layout_last_changed_on`, then the component array itself
            # has been modified. So we should force a re-render of the card.
            self._last_layout_change = self.components.layout_last_changed_on
        else:
            self._card_proc("refresh")

    @property
    def components(self):
        return self._components

    def _warning(self, message):
        msg = "[@card WARNING] %s" % message
        self._logger(msg, timestamp=False, bad=True)

    def _get_latest_data(self, final=False, mode=None):
        """
        This function returns the data object that is passed down to :
        - `MetaflowCard.render_runtime`
        - `MetaflowCard.refresh`
        - `MetaflowCard.reload_content_token`

        The return value of this function contains all the necessary state information for Metaflow Cards to make decisions on the following:
        1. What components are rendered
        2. Should the card be reloaded on the UI
        3. What data to pass down to the card.

        Parameters
        ----------
        final : bool, optional
            If True, it implies that the final "rendering" sequence is taking place (which involves calling a `render` and a `refresh` function.)
            When final is set the `render_seq` is set to "final" so that the reload token in the card is set to final
            and the card is not reloaded again on the user interface.
        mode : str
            This parameter is passed down to the object returned by this function. Can be one of `render_runtime` / `refresh` / `render`

        Returns
        -------
        dict
            A dictionary of the form :
            ```python
            {
                "user": user_data, # any passed to `current.card.refresh` function
                "components": component_dict, # all rendered REALTIME_UPDATABLE components
                "render_seq": seq,
                # `render_seq` is a counter that is incremented every time `render_runtime` is called.
                # If a metaflow card has a RELOAD_POLICY_ALWAYS set then the reload token will be set to this value
                # so that the card reload on the UI everytime `render_runtime` is called.
                "component_update_ts": self.components.layout_last_changed_on,
                # `component_update_ts` is the timestamp of the last time the component array was modified.
                # `component_update_ts` can get used by the `reload_content_token` to make decisions on weather to
                # reload the card on the UI when component array has changed.
                "mode": mode,
            }
            ```
        """
        seq = "final" if final else self._render_seq
        # Extract all the runtime-updatable components as a dictionary
        component_dict = {}
        for component in self._components._realtime_updateable_components():
            rendered_comp = _render_card_component(component)
            if rendered_comp is not None:
                component_dict.update({component.component_id: rendered_comp})

        # Verify _latest_user_data is json serializable
        user_data = {}
        if self._latest_user_data is not None and not _object_is_json_serializable(
            self._latest_user_data
        ):
            self._warning(
                "Data provided to `refresh` is not JSON serializable. It will be ignored."
            )
        else:
            user_data = self._latest_user_data

        return {
            "user": user_data,
            "components": component_dict,
            "render_seq": seq,
            "component_update_ts": self.components.layout_last_changed_on,
            "mode": mode,
        }

    def __iter__(self):
        return iter(self._components)


class CardComponentCollector:
    """
    This class helps collect `MetaflowCardComponent`s during runtime execution

    ### Usage with `current`
    `current.card` is of type `CardComponentCollector`

    ### Main Usage TLDR
    - [x] `current.card.append` customizes the default editable card.
    - [x] Only one card can be default editable in a step.
    - [x] The card class must have `ALLOW_USER_COMPONENTS=True` to be considered default editable.
        - [x] Classes with `ALLOW_USER_COMPONENTS=False` are never default editable.
    - [x] The user can specify an `id` argument to a card, in which case the card is editable through `current.card[id].append`.
        - [x] A card with an id can be also default editable, if there are no other cards that are eligible to be default editable.
    - [x] If multiple default-editable cards exist but only one card doesn't have an id, the card without an id is considered to be default editable.
    - [x] If we can't resolve a single default editable card through the above rules, `current.card`.append calls show a warning but the call doesn't fail.
    - [x] A card that is not default editable can be still edited through:
        - [x] its `current.card['myid']`
        - [x] by looking it up by its type, e.g. `current.card.get(type='pytorch')`.
    """

    def __init__(self, logger=None, card_creator=None):
        from metaflow.metaflow_config import CARD_NO_WARNING

        self._card_component_store = (
            # Each key in the dictionary is the UUID of an individual card.
            # value is of type `CardComponentManager`, holding a list of MetaflowCardComponents for that particular card
            {}
        )
        self._cards_meta = (
            {}
        )  # a `dict` of (card_uuid, `dict)` holding all metadata about all @card decorators on the `current` @step.
        self._card_id_map = {}  # card_id to uuid map for all cards with ids
        self._logger = logger
        self._card_creator = card_creator
        # `self._default_editable_card` holds the uuid of the card that is default editable. This card has access to `append`/`extend` methods of `self`
        self._default_editable_card = None
        self._warned_once = {
            "__getitem__": {},
            "append": False,
            "extend": False,
            "update": False,
            "update_no_id": False,
        }
        self._no_warnings = True if CARD_NO_WARNING else False

    @staticmethod
    def create_uuid():
        return str(uuid.uuid4()).replace("-", "")

    def _log(self, *args, **kwargs):
        if self._logger:
            self._logger(*args, **kwargs)

    def _add_card(
        self,
        card_type,
        card_id,
        decorator_attributes,
        card_options,
        editable=False,
        customize=False,
        suppress_warnings=False,
        runtime_card=False,
        refresh_interval=5,
    ):
        """
        This function helps collect cards from all the card decorators.
        As `current.card` is a singleton this function is called by all @card decorators over a @step to add editable cards.

        ## Parameters

            - `card_type` (str) : value of the associated `MetaflowCard.type`
            - `card_id` (str) : `id` argument provided at top of decorator
            - `editable` (bool) : this corresponds to the value of `MetaflowCard.ALLOW_USER_COMPONENTS` for that `card_type`
            - `customize` (bool) : This argument is reserved for a single @card decorator per @step.
                - An `editable` card with `customize=True` gets precedence to be set as default editable card.
                - A default editable card is the card which can be access via the `append` and `extend` methods.
        """
        card_uuid = self.create_uuid()
        card_metadata = dict(
            type=card_type,
            uuid=card_uuid,
            card_id=card_id,
            editable=editable,
            customize=customize,
            suppress_warnings=suppress_warnings,
            runtime_card=runtime_card,
            decorator_attributes=decorator_attributes,
            card_options=card_options,
            refresh_interval=refresh_interval,
        )
        self._cards_meta[card_uuid] = card_metadata
        self._card_component_store[card_uuid] = CardComponentManager(
            card_uuid,
            decorator_attributes,
            self._card_creator,
            components=None,
            logger=self._logger,
            no_warnings=self._no_warnings,
            user_set_card_id=card_id,
            runtime_card=runtime_card,
            card_options=card_options,
            refresh_interval=refresh_interval,
        )
        return card_metadata

    def _warning(self, message):
        msg = "[@card WARNING] %s" % message
        self._log(msg, timestamp=False, bad=True)

    def _add_warning_to_cards(self, warn_msg):
        if self._no_warnings:
            return
        for card_id in self._card_component_store:
            if not self._cards_meta[card_id]["suppress_warnings"]:
                self._card_component_store[card_id].append(WarningComponent(warn_msg))

    def get(self, type=None):
        """`get`
        gets all the components arrays for a card `type`.
        Since one `@step` can have many `@card` decorators, many decorators can have the same type. That is why this function returns a list of lists.

        Args:
            type ([str], optional): `type` of MetaflowCard. Defaults to None.

        Returns: will return empty `list` if `type` is None or not found
            List[List[MetaflowCardComponent]]
        """
        card_type = type
        card_uuids = [
            card_meta["uuid"]
            for card_meta in self._cards_meta.values()
            if card_meta["type"] == card_type
        ]
        return [self._card_component_store[uuid] for uuid in card_uuids]

    def _finalize(self):
        """
        The `_finalize` function is called once the last @card decorator calls `step_init`. Calling this function makes `current.card` ready for usage inside `@step` code.
        This function's works two parts :
        1. Resolving `self._default_editable_card`.
                - The `self._default_editable_card` holds the uuid of the card that will have access to the `append`/`extend` methods.
        2. Resolving edge cases where @card `id` argument may be `None` or have a duplicate `id` when there are more than one editable cards.
        3. Resolving the `self._default_editable_card` to the card with the`customize=True` argument.
        """
        all_card_meta = list(self._cards_meta.values())
        for c in all_card_meta:
            ct = get_card_class(c["type"])
            c["exists"] = False
            if ct is not None:
                c["exists"] = True

        # If a card has `customize=True` and is not editable then it will not be considered default editable.
        editable_cards_meta = [c for c in all_card_meta if c["editable"]]

        if len(editable_cards_meta) == 0:
            return

        # Create the `self._card_id_map` lookup table which maps card `id` to `uuid`.
        # This table has access to all cards with `id`s set to them.
        card_ids = []
        for card_meta in all_card_meta:
            if card_meta["card_id"] is not None:
                self._card_id_map[card_meta["card_id"]] = card_meta["uuid"]
                card_ids.append(card_meta["card_id"])

        # If there is only one editable card then this card becomes `self._default_editable_card`
        if len(editable_cards_meta) == 1:
            self._default_editable_card = editable_cards_meta[0]["uuid"]
            return

        # Segregate cards which have id as none and those which don't.
        not_none_id_cards = [c for c in editable_cards_meta if c["card_id"] is not None]
        none_id_cards = [c for c in editable_cards_meta if c["card_id"] is None]

        # If there is only 1 card with id set to None then we can use that as the default card.
        if len(none_id_cards) == 1:
            self._default_editable_card = none_id_cards[0]["uuid"]

        # If the size of the set of ids is not equal to total number of cards with ids then warn the user that we cannot disambiguate
        # so `current.card['my_card_id']` won't work.
        id_set = set(card_ids)
        if len(card_ids) != len(id_set):
            non_unique_ids = [
                idx
                for idx in id_set
                if len(list(filter(lambda x: x["card_id"] == idx, not_none_id_cards)))
                > 1
            ]
            nui = ", ".join(non_unique_ids)
            # throw a warning that decorators have non-unique Ids
            self._warning(
                (
                    "Multiple `@card` decorator have been annotated with duplicate ids : %s. "
                    "`current.card['%s']` will not work"
                )
                % (nui, non_unique_ids[0])
            )

            # remove the non unique ids from the `self._card_id_map`
            for idx in non_unique_ids:
                del self._card_id_map[idx]

        # if a @card has `customize=True` in the arguments then there should only be one @card with `customize=True`. This @card will be the _default_editable_card
        customize_cards = [c for c in editable_cards_meta if c["customize"]]
        if len(customize_cards) > 1:
            self._warning(
                (
                    "Multiple @card decorators have `customize=True`. "
                    "Only one @card per @step can have `customize=True`. "
                    "`current.card.append` will ignore all decorators marked `customize=True`."
                )
            )
        elif len(customize_cards) == 1:
            # since `editable_cards_meta` hold only `editable=True` by default we can just set this card here.
            self._default_editable_card = customize_cards[0]["uuid"]

    def __getitem__(self, key):
        """
        Choose a specific card for manipulation.

        When multiple @card decorators are present, you can add an
        `ID` to distinguish between them, `@card(id=ID)`. This allows you
        to add components to a specific card like this:
        ```
        current.card[ID].append(component)
        ```

        Parameters
        ----------
        key : str
            Card ID.

        Returns
        -------
        CardComponentManager
            An object with `append` and `extend` calls which allow you to
            add components to the chosen card.
        """
        if key in self._card_id_map:
            card_uuid = self._card_id_map[key]
            return self._card_component_store[card_uuid]
        if key not in self._warned_once["__getitem__"]:
            _warn_msg = [
                "`current.card['%s']` is not present. Please set the `id` argument in @card to '%s' to access `current.card['%s']`."
                % (key, key, key),
                "`current.card['%s']` will return an empty `list` which is not referenced to `current.card` object."
                % (key),
            ]
            self._warning(" ".join(_warn_msg))
            self._add_warning_to_cards("\n".join(_warn_msg))
            self._warned_once["__getitem__"][key] = True

        return []

    def __setitem__(self, key, value):
        """
        Specify components of the chosen card.

        Instead of adding components to a card individually with `current.card[ID].append(component)`,
        use this method to assign a list of components to a card, replacing the existing components:
        ```
        current.card[ID] = [FirstComponent, SecondComponent]
        ```

        Parameters
        ----------
        key: str
            Card ID.

        value: List[MetaflowCardComponent]
            List of card components to assign to this card.
        """
        if key in self._card_id_map:
            card_uuid = self._card_id_map[key]
            if not isinstance(value, list):
                _warning_msg = (
                    "`current.card['%s']` not set. `current.card['%s']` should be a `list` of `MetaflowCardComponent`."
                    % (key, key)
                )
                self._warning(_warning_msg)
                return
            self._card_component_store[card_uuid] = CardComponentManager(
                card_uuid,
                self._cards_meta[card_uuid]["decorator_attributes"],
                self._card_creator,
                components=value,
                logger=self._logger,
                no_warnings=self._no_warnings,
                user_set_card_id=key,
                card_options=self._cards_meta[card_uuid]["card_options"],
                runtime_card=self._cards_meta[card_uuid]["runtime_card"],
                refresh_interval=self._cards_meta[card_uuid]["refresh_interval"],
            )
            return

        self._warning(
            "`current.card['%s']` is not present. Please set the `id` argument in @card to '%s' to access `current.card['%s']`. "
            % (key, key, key)
        )

    def append(self, component, id=None):
        """
        Appends a component to the current card.

        Parameters
        ----------
        component : MetaflowCardComponent
            Card component to add to this card.
        """
        if self._default_editable_card is None:
            if (
                len(self._card_component_store) == 1
            ):  # if there is one card which is not the _default_editable_card then the card is not editable
                card_type = list(self._cards_meta.values())[0]["type"]
                if list(self._cards_meta.values())[0]["exists"]:
                    _crdwr = "Card of type `%s` is not an editable card." % card_type
                    _endwr = (
                        "Please use an editable card."  # todo : link to documentation
                    )
                else:
                    _crdwr = "Card of type `%s` doesn't exist." % card_type
                    _endwr = "Please use a card `type` which exits."  # todo : link to documentation

                _warning_msg = [
                    _crdwr,
                    "Component will not be appended and `current.card.append` will not work for any call during this runtime execution.",
                    _endwr,
                ]
            else:
                _warning_msg = [
                    "`current.card.append` cannot disambiguate between multiple editable cards.",
                    "Component will not be appended and `current.card.append` will not work for any call during this runtime execution.",
                    "To fix this set the `id` argument in all @card's when using multiple @card decorators over a single @step. ",  # todo : Add Link to documentation
                ]

            if not self._warned_once["append"]:
                self._warning(" ".join(_warning_msg))
                self._add_warning_to_cards("\n".join(_warning_msg))
                self._warned_once["append"] = True

            return
        self._card_component_store[self._default_editable_card].append(component, id=id)

    def extend(self, components):
        """
        Appends many components to the current card.

        Parameters
        ----------
        component : Iterator[MetaflowCardComponent]
            Card components to add to this card.
        """
        if self._default_editable_card is None:
            # if there is one card which is not the _default_editable_card then the card is not editable
            if len(self._card_component_store) == 1:
                card_type = list(self._cards_meta.values())[0]["type"]
                _warning_msg = [
                    "Card of type `%s` is not an editable card." % card_type,
                    "Components list will not be extended and `current.card.extend` will not work for any call during this runtime execution.",
                    "Please use an editable card",  # todo : link to documentation
                ]
            else:
                _warning_msg = [
                    "`current.card.extend` cannot disambiguate between multiple @card decorators.",
                    "Components list will not be extended and `current.card.extend` will not work for any call during this runtime execution.",
                    "To fix this set the `id` argument in all @card when using multiple @card decorators over a single @step.",  # todo : Add Link to documentation
                ]
            if not self._warned_once["extend"]:
                self._warning(" ".join(_warning_msg))
                self._add_warning_to_cards("\n".join(_warning_msg))
                self._warned_once["extend"] = True

            return

        self._card_component_store[self._default_editable_card].extend(components)

    @property
    def components(self):
        # FIXME: document
        if self._default_editable_card is None:
            if len(self._card_component_store) == 1:
                card_type = list(self._cards_meta.values())[0]["type"]
                _warning_msg = [
                    "Card of type `%s` is not an editable card." % card_type,
                    "Components list will not be updated and `current.card.components` will not work for any call during this runtime execution.",
                    "Please use an editable card",  # todo : link to documentation
                ]
            else:
                _warning_msg = [
                    "`current.card.components` cannot disambiguate between multiple @card decorators.",
                    "Components list will not be accessible and `current.card.components` will not work for any call during this runtime execution.",
                    "To fix this set the `id` argument in all @card when using multiple @card decorators over a single @step and reference `current.card[ID].components`",  # todo : Add Link to documentation
                    "to update/access the appropriate card component.",
                ]
            if not self._warned_once["components"]:
                self._warning(" ".join(_warning_msg))
                self._warned_once["components"] = True
            return

        return self._card_component_store[self._default_editable_card].components

    def clear(self):
        # FIXME: document
        if self._default_editable_card is not None:
            self._card_component_store[self._default_editable_card].clear()

    def refresh(self, *args, **kwargs):
        # FIXME: document
        if self._default_editable_card is not None:
            self._card_component_store[self._default_editable_card].refresh(
                *args, **kwargs
            )

    def _get_latest_data(self, card_uuid, final=False, mode=None):
        """
        Returns latest data so it can be used in the final render() call
        """
        return self._card_component_store[card_uuid]._get_latest_data(
            final=final, mode=mode
        )

    def _serialize_components(self, card_uuid):
        """
        This method renders components present in a card to strings/json.
        Components exposed by metaflow ensure that they render safely. If components
        don't render safely then we don't add them to the final list of serialized functions
        """
        serialized_components = []
        if card_uuid not in self._card_component_store:
            return []
        has_user_components = any(
            [
                issubclass(type(component), UserComponent)
                for component in self._card_component_store[card_uuid]
            ]
        )
        for component in self._card_component_store[card_uuid]:
            rendered_obj = _render_card_component(component)
            if rendered_obj is None:
                continue
            serialized_components.append(rendered_obj)
        if has_user_components and len(serialized_components) > 0:
            serialized_components = [
                SectionComponent(contents=serialized_components).render()
            ]
        return serialized_components


def _render_card_component(component):
    if not _component_is_valid(component):
        return None
    try:
        rendered_obj = component.render()
    except:
        return None
    else:
        if not (type(rendered_obj) == str or type(rendered_obj) == dict):
            return None
        else:
            # Since `UserComponent`s are safely_rendered using render_tools.py
            # we don't need to check JSON serialization as @render_tools.render_safely
            # decorator ensures this check so there is no need to re-serialize
            if not issubclass(type(component), UserComponent):
                try:  # check if rendered object is json serializable.
                    json.dumps(rendered_obj)
                except (TypeError, OverflowError) as e:
                    return None
        return rendered_obj



import inspect
import json
import collections.abc
import copy
import os
import re

from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TYPE_CHECKING,
    Union,
)


from ..exception import MetaflowException

from ..parameters import (
    Parameter,
    ParameterContext,
    current_flow,
)

if TYPE_CHECKING:
    from metaflow import FlowSpec

# _tracefunc_depth = 0


# def tracefunc(func):
#     """Decorates a function to show its trace."""

#     @functools.wraps(func)
#     def tracefunc_closure(*args, **kwargs):
#         global _tracefunc_depth
#         """The closure."""
#         print(f"{_tracefunc_depth}: {func.__name__}(args={args}, kwargs={kwargs})")
#         _tracefunc_depth += 1
#         result = func(*args, **kwargs)
#         _tracefunc_depth -= 1
#         print(f"{_tracefunc_depth} => {result}")
#         return result

#     return tracefunc_closure

ID_PATTERN = re.compile(r"^[a-zA-Z_][a-zA-Z0-9_]*$")

UNPACK_KEY = "_unpacked_delayed_"


def dump_config_values(flow: "FlowSpec"):
    from ..flowspec import _FlowState  # Prevent circular import

    configs = flow._flow_state.get(_FlowState.CONFIGS)
    if configs:
        return {"user_configs": configs}
    return {}


class ConfigValue(collections.abc.Mapping, dict):
    """
    ConfigValue is a thin wrapper around an arbitrarily nested dictionary-like
    configuration object. It allows you to access elements of this nested structure
    using either a "." notation or a [] notation. As an example, if your configuration
    object is:
    {"foo": {"bar": 42}}
    you can access the value 42 using either config["foo"]["bar"] or config.foo.bar.

    All "keys"" need to be valid Python identifiers
    """

    # Thin wrapper to allow configuration values to be accessed using a "." notation
    # as well as a [] notation.

    # We inherit from dict to allow the isinstanceof check to work easily and also
    # to provide a simple json dumps functionality.

    def __init__(self, data: Union["ConfigValue", Dict[str, Any]]):
        self._data = {k: self._construct(v) for k, v in data.items()}

        # Enable json dumps
        dict.__init__(self, self._data)

    @classmethod
    def fromkeys(cls, iterable: Iterable, value: Any = None) -> "ConfigValue":
        """
        Creates a new ConfigValue object from the given iterable and value.

        Parameters
        ----------
        iterable : Iterable
            Iterable to create the ConfigValue from.
        value : Any, optional
            Value to set for each key in the iterable.

        Returns
        -------
        ConfigValue
            A new ConfigValue object.
        """
        return cls(dict.fromkeys(iterable, value))

    def to_dict(self) -> Dict[Any, Any]:
        """
        Returns a dictionary representation of this configuration object.

        Returns
        -------
        Dict[Any, Any]
            Dictionary equivalent of this configuration object.
        """
        return self._to_dict(self._data)

    def copy(self) -> "ConfigValue":
        return self.__copy__()

    def clear(self) -> None:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def update(self, *args, **kwargs) -> None:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def setdefault(self, key: Any, default: Any = None) -> Any:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def pop(self, key: Any, default: Any = None) -> Any:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def popitem(self) -> Tuple[Any, Any]:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def __getattr__(self, key: str) -> Any:
        """
        Access an element of this configuration

        Parameters
        ----------
        key : str
            Element to access

        Returns
        -------
        Any
            Element of the configuration
        """
        if key == "_data":
            # Called during unpickling. Special case to not run into infinite loop
            # below.
            raise AttributeError(key)

        if key in self._data:
            return self[key]
        raise AttributeError(key)

    def __setattr__(self, name: str, value: Any) -> None:
        # Prevent configuration modification
        if name == "_data":
            return super().__setattr__(name, value)
        raise TypeError("ConfigValue is immutable")

    def __getitem__(self, key: Any) -> Any:
        """
        Access an element of this configuration

        Parameters
        ----------
        key : Any
            Element to access

        Returns
        -------
        Any
            Element of the configuration
        """
        return self._data[key]

    def __setitem__(self, key: Any, value: Any) -> None:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def __delattr__(self, key) -> None:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def __delitem__(self, key: Any) -> None:
        # Prevent configuration modification
        raise TypeError("ConfigValue is immutable")

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Iterator:
        return iter(self._data)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ConfigValue):
            return self._data == other._data
        if isinstance(other, dict):
            return self._data == other
        return False

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __copy__(self) -> "ConfigValue":
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update({k: copy.copy(v) for k, v in self.__dict__.items()})
        return result

    def __repr__(self) -> str:
        return repr(self._data)

    def __str__(self) -> str:
        return str(self._data)

    def __dir__(self) -> Iterable[str]:
        return dir(type(self)) + [k for k in self._data.keys() if ID_PATTERN.match(k)]

    def __contains__(self, key: Any) -> bool:
        try:
            self[key]
        except KeyError:
            return False
        return True

    def keys(self):
        """
        Returns the keys of this configuration object.

        Returns
        -------
        Any
            Keys of this configuration object.
        """
        return self._data.keys()

    @classmethod
    def _construct(cls, obj: Any) -> Any:
        # Internal method to construct a ConfigValue so that all mappings internally
        # are also converted to ConfigValue
        if isinstance(obj, ConfigValue):
            v = obj
        elif isinstance(obj, collections.abc.Mapping):
            v = ConfigValue({k: cls._construct(v) for k, v in obj.items()})
        elif isinstance(obj, (list, tuple, set)):
            v = type(obj)([cls._construct(x) for x in obj])
        else:
            v = obj
        return v

    @classmethod
    def _to_dict(cls, obj: Any) -> Any:
        # Internal method to convert all nested mappings to dicts
        if isinstance(obj, collections.abc.Mapping):
            v = {k: cls._to_dict(v) for k, v in obj.items()}
        elif isinstance(obj, (list, tuple, set)):
            v = type(obj)([cls._to_dict(x) for x in obj])
        else:
            v = obj
        return v

    def __reduce__(self):
        return (self.__class__, (self.to_dict(),))


class DelayEvaluator(collections.abc.Mapping):
    """
    Small wrapper that allows the evaluation of a Config() value in a delayed manner.
    This is used when we want to use config.* values in decorators for example.

    It also allows the following "delayed" access on an obj that is a DelayEvaluation
      - obj.x.y.z (ie: accessing members of DelayEvaluator; accesses will be delayed until
        the DelayEvaluator is evaluated)
      - **obj (ie: unpacking the DelayEvaluator as a dictionary). Note that this requires
        special handling in whatever this is being unpacked into, specifically the handling
        of _unpacked_delayed_*
    """

    def __init__(self, ex: str, saved_globals: Optional[Dict[str, Any]] = None):
        self._config_expr = ex
        self._globals = saved_globals
        if ID_PATTERN.match(self._config_expr):
            # This is a variable only so allow things like config_expr("config").var
            self._is_var_only = True
            self._access = []
        else:
            self._is_var_only = False
            self._access = None
        self._cached_expr = None

    def __copy__(self):
        c = DelayEvaluator(self._config_expr)
        c._access = self._access.copy() if self._access is not None else None
        # Globals are not copied -- always kept as a reference
        return c

    def __deepcopy__(self, memo):
        c = DelayEvaluator(self._config_expr)
        c._access = (
            copy.deepcopy(self._access, memo) if self._access is not None else None
        )
        # Globals are not copied -- always kept as a reference
        return c

    def __iter__(self):
        yield "%s%d" % (UNPACK_KEY, id(self))

    def __getitem__(self, key):
        if isinstance(key, str) and key == "%s%d" % (UNPACK_KEY, id(self)):
            return self
        if self._access is None:
            raise KeyError(key)

        # Make a copy so that we can support something like
        # foo = delay_evaluator["blah"]
        # bar = delay_evaluator["baz"]
        # and don't end up with a access list that contains both "blah" and "baz"
        c = self.__copy__()
        c._access.append(key)
        c._cached_expr = None
        return c

    def __len__(self):
        return 1

    def __getattr__(self, name):
        if self._access is None:
            raise AttributeError(name)
        c = self.__copy__()
        c._access.append(name)
        c._cached_expr = None
        return c

    def __call__(self, ctx=None, deploy_time=False):
        from ..flowspec import _FlowState  # Prevent circular import

        # Two additional arguments are only used by DeployTimeField which will call
        # this function with those two additional arguments. They are ignored.
        flow_cls = getattr(current_flow, "flow_cls", None)
        if flow_cls is None:
            # We are not executing inside a flow (ie: not the CLI)
            raise MetaflowException(
                "Config object can only be used directly in the FlowSpec defining them "
                "(or their flow decorators)."
            )
        if self._cached_expr is not None:
            to_eval_expr = self._cached_expr
        elif self._access is not None:
            # Build the final expression by adding all the fields in access as . fields
            access_list = [self._config_expr]
            for a in self._access:
                if isinstance(a, str):
                    access_list.append(a)
                elif isinstance(a, DelayEvaluator):
                    # Supports things like config[other_config.selector].var
                    access_list.append(a())
                else:
                    raise MetaflowException(
                        "Field '%s' of type '%s' is not supported" % (str(a), type(a))
                    )
            to_eval_expr = self._cached_expr = ".".join(access_list)
        else:
            to_eval_expr = self._cached_expr = self._config_expr
        # Evaluate the expression setting the config values as local variables
        try:
            return eval(
                to_eval_expr,
                self._globals or globals(),
                {
                    k: ConfigValue(v)
                    for k, v in flow_cls._flow_state.get(_FlowState.CONFIGS, {}).items()
                },
            )
        except NameError as e:
            raise MetaflowException(
                "Config expression '%s' could not be evaluated: %s"
                % (to_eval_expr, str(e))
            ) from e


def config_expr(expr: str) -> DelayEvaluator:
    """
    Function to allow you to use an expression involving a config parameter in
    places where it may not be directory accessible or if you want a more complicated
    expression than just a single variable.

    You can use it as follows:
      - When the config is not directly accessible:

            @project(name=config_expr("config").project.name)
            class MyFlow(FlowSpec):
                config = Config("config")
                ...
      - When you want a more complex expression:
            class MyFlow(FlowSpec):
                config = Config("config")

                @environment(vars={"foo": config_expr("config.bar.baz.lower()")})
                @step
                def start(self):
                    ...

    Parameters
    ----------
    expr : str
        Expression using the config values.
    """
    # Get globals where the expression is defined so that the user can use
    # something like `config_expr("my_func()")` in the expression.
    parent_globals = inspect.currentframe().f_back.f_globals
    return DelayEvaluator(expr, saved_globals=parent_globals)


class Config(Parameter, collections.abc.Mapping):
    """
    Includes a configuration for this flow.

    `Config` is a special type of `Parameter` but differs in a few key areas:
      - it is immutable and determined at deploy time (or prior to running if not deploying
        to a scheduler)
      - as such, it can be used anywhere in your code including in Metaflow decorators

    The value of the configuration is determines as follows:
      - use the user-provided file path or value. It is an error to provide both
      - if none are present:
        - if a default file path (default) is provided, attempt to read this file
            - if the file is present, use that value. Note that the file will be used
              even if it has an invalid syntax
            - if the file is not present, and a default value is present, use that
      - if still None and is required, this is an error.

    Parameters
    ----------
    name : str
        User-visible configuration name.
    default : Union[str, Callable[[ParameterContext], str], optional, default None
        Default path from where to read this configuration. A function implies that the
        value will be computed using that function.
        You can only specify default or default_value, not both.
    default_value : Union[str, Dict[str, Any], Callable[[ParameterContext, Union[str, Dict[str, Any]]], Any], optional, default None
        Default value for the parameter. A function
        implies that the value will be computed using that function.
        You can only specify default or default_value, not both.
    help : str, optional, default None
        Help text to show in `run --help`.
    required : bool, optional, default None
        Require that the user specifies a value for the configuration. Note that if
        a default or default_value is provided, the required flag is ignored.
        A value of None is equivalent to False.
    parser : Union[str, Callable[[str], Dict[Any, Any]]], optional, default None
        If a callable, it is a function that can parse the configuration string
        into an arbitrarily nested dictionary. If a string, the string should refer to
        a function (like "my_parser_package.my_parser.my_parser_function") which should
        be able to parse the configuration string into an arbitrarily nested dictionary.
        If the name starts with a ".", it is assumed to be relative to "metaflow".
    show_default : bool, default True
        If True, show the default value in the help text.
    """

    IS_CONFIG_PARAMETER = True

    def __init__(
        self,
        name: str,
        default: Optional[Union[str, Callable[[ParameterContext], str]]] = None,
        default_value: Optional[
            Union[
                str,
                Dict[str, Any],
                Callable[[ParameterContext], Union[str, Dict[str, Any]]],
            ]
        ] = None,
        help: Optional[str] = None,
        required: Optional[bool] = None,
        parser: Optional[Union[str, Callable[[str], Dict[Any, Any]]]] = None,
        **kwargs: Dict[str, str]
    ):
        if default is not None and default_value is not None:
            raise MetaflowException(
                "For config '%s', you can only specify default or default_value, not both"
                % name
            )
        self._default_is_file = default is not None
        kwargs["default"] = default if default is not None else default_value
        super(Config, self).__init__(
            name, required=required, help=help, type=str, **kwargs
        )
        super(Config, self).init()

        if isinstance(kwargs.get("default", None), str):
            kwargs["default"] = json.dumps(kwargs["default"])
        self.parser = parser
        self._computed_value = None

        self._delayed_evaluator = None

    def load_parameter(self, v):
        if v is None:
            return None
        return ConfigValue(v)

    def _store_value(self, v: Any) -> None:
        self._computed_value = v

    def _init_delayed_evaluator(self) -> None:
        if self._delayed_evaluator is None:
            self._delayed_evaluator = DelayEvaluator(self.name.lower())

    # Support <config>.<var> syntax
    def __getattr__(self, name):
        # Need to return a new DelayEvaluator everytime because the evaluator will
        # contain the "path" (ie: .name) and can be further accessed.
        return getattr(DelayEvaluator(self.name.lower()), name)

    # Next three methods are to implement mapping to support **<config> syntax. We
    # need to be careful, however, to also support a regular `config["key"]` syntax
    # which calls into `__getitem__` and therefore behaves like __getattr__ above.
    def __iter__(self):
        self._init_delayed_evaluator()
        yield from self._delayed_evaluator

    def __len__(self):
        self._init_delayed_evaluator()
        return len(self._delayed_evaluator)

    def __getitem__(self, key):
        self._init_delayed_evaluator()
        if isinstance(key, str) and key.startswith(UNPACK_KEY):
            return self._delayed_evaluator[key]
        return DelayEvaluator(self.name.lower())[key]


def resolve_delayed_evaluator(
    v: Any, ignore_errors: bool = False, to_dict: bool = False
) -> Any:
    # NOTE: We don't ignore errors in downstream calls because we want to have either
    # all or nothing for the top-level call by the user.
    try:
        if isinstance(v, DelayEvaluator):
            to_return = v()
            if to_dict and isinstance(to_return, ConfigValue):
                to_return = to_return.to_dict()
            return to_return
        if isinstance(v, dict):
            return {
                resolve_delayed_evaluator(
                    k, to_dict=to_dict
                ): resolve_delayed_evaluator(v, to_dict=to_dict)
                for k, v in v.items()
            }
        if isinstance(v, list):
            return [resolve_delayed_evaluator(x, to_dict=to_dict) for x in v]
        if isinstance(v, tuple):
            return tuple(resolve_delayed_evaluator(x, to_dict=to_dict) for x in v)
        if isinstance(v, set):
            return {resolve_delayed_evaluator(x, to_dict=to_dict) for x in v}
        return v
    except Exception as e:
        if ignore_errors:
            # Assumption is that default value of None is always allowed.
            # This code path is *only* used when evaluating Parameters AND they
            # use configs in their attributes AND the runner/deployer is being used
            # AND CLICK_API_PROCESS_CONFIG is False. In those cases, all attributes in
            # Parameter can be set to None except for required and show_default
            # and even in those cases, a wrong value will have very limited consequence.
            return None
        raise e


def unpack_delayed_evaluator(
    to_unpack: Dict[str, Any], ignore_errors: bool = False
) -> Tuple[Dict[str, Any], List[str]]:
    result = {}
    new_keys = []
    for k, v in to_unpack.items():
        if not isinstance(k, str) or not k.startswith(UNPACK_KEY):
            result[k] = v
        else:
            # k.startswith(UNPACK_KEY)
            try:
                new_vals = resolve_delayed_evaluator(v, to_dict=True)
                new_keys.extend(new_vals.keys())
                result.update(new_vals)
            except Exception as e:
                if ignore_errors:
                    continue
                raise e
    return result, new_keys
